/****************************************************************************
 *
 *	Copyright (c) 1999-2008, Watanabe Lab, School of Media Science,
 *	Tokyo University of Technology, All rights reserved.
 *
 *	Redistribution and use in source and binary forms,
 *	with or without modification, are permitted provided that the
 *	following conditions are met:
 *
 *		- Redistributions of source code must retain the above
 *			copyright notice, this list of conditions and the
 *			following disclaimer.
 *
 *		- Redistributions in binary form must reproduce the above
 *			copyright notice, this list of conditions and the
 *			following disclaimer in the documentation and/or
 *			other materials provided with the distribution.
 *
 *		- Neither the name of the copyright holders nor the names
 *			of its contributors may be used to endorse or promote
 *			products derived from this software without specific
 *			prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *	FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *	COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 *	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE. 
 *
 ****************************************************************************/
/****************************************************************************
 *
 *	Copyright (c) 1999-2008, Watanabe Lab, School of Media Science,
 *	Tokyo University of Technology, All rights reserved.
 *
 *	本ソフトウェアおよびソースコードのライセンスは、基本的に
 *	「修正 BSD ライセンス」に従います。以下にその詳細を記します。
 *
 *	ソースコード形式かバイナリ形式か、変更するかしないかを問わず、
 *	以下の条件を満たす場合に限り、再頒布および使用が許可されます。
 *
 *	- ソースコードを再頒布する場合、上記の著作権表示、本条件一覧、
 *		および下記免責条項を含めること。
 *
 *	- バイナリ形式で再頒布する場合、頒布物に付属のドキュメント等の
 *		資料に、上記の著作権表示、本条件一覧、および下記免責条項を
 *		含めること。
 *
 *	- 書面による特別の許可なしに、本ソフトウェアから派生した製品の
 *		宣伝または販売促進に、本ソフトウェアの著作権者の名前または
 *		コントリビューターの名前を使用してはならない。
 *
 *	本ソフトウェアは、著作権者およびコントリビューターによって「現
 *	状のまま」提供されており、明示黙示を問わず、商業的な使用可能性、
 *	および特定の目的に対する適合性に関す暗黙の保証も含め、またそれ
 *	に限定されない、いかなる保証もないものとします。著作権者もコン
 *	トリビューターも、事由のいかんを問わず、損害発生の原因いかんを
 *	問わず、かつ責任の根拠が契約であるか厳格責任であるか(過失その
 *	他の)不法行為であるかを問わず、仮にそのような損害が発生する可
 *	能性を知らされていたとしても、本ソフトウェアの使用によって発生
 *	した(代替品または代用サービスの調達、使用の喪失、データの喪失、
 *	利益の喪失、業務の中断も含め、またそれに限定されない)直接損害、
 *	間接損害、偶発的な損害、特別損害、懲罰的損害、または結果損害に
 *	ついて、一切責任を負わないものとします。
 *
 ****************************************************************************/
#ifndef __FK_PARTICLE_HEADER__
#define __FK_PARTICLE_HEADER__

#include <FK/Base.h>
#include <FK/Vector.h>
#include <FK/Material.h>

class fk_Point;
class fk_IDAdmin;
class fk_Shape;

class fk_Particle: public fk_BaseObject {

 private:
	int				id;			// 頂点 ID 
	fk_Point		*base;		// 全体位相 
	unsigned int	count;		// 年齢 
	fk_Vector		velocity;	// 速度 
	fk_Vector		accel;		// 加速度 

 public:
	fk_Particle(fk_Point *, int);
	~fk_Particle();

	// 初期化 
	void			init(void);

	// ID 取得 
	int				getID(void) const;

	// 年齢取得 
	unsigned int	getCount(void) const;

	// 位置取得設定 
	fk_Vector		getPosition(void) const;
	void			setPosition(const fk_Vector &);
	void			setPosition(double, double, double);

	// 速度取得設定 
	fk_Vector		getVelocity(void) const;
	void			setVelocity(const fk_Vector &);
	void			setVelocity(double, double, double);

	// 加速度取得設定 
	fk_Vector		getAccel(void) const;
	void			setAccel(const fk_Vector &);
	void			setAccel(double, double, double);

	// 色 ID 取得設定 
	int				getColorID(void) const;
	void			setColorID(int);

	// 色モード取得設定 
	bool			getDrawMode(void) const;
	void			setDrawMode(bool);

	// 動作 
	void			handle(void);
};

class fk_ParticleSet: public fk_BaseObject {

 private:
	vector<fk_Particle *>	pSet;
	fk_IDAdmin				*pAdmin;
	fk_Point				*point;
	unsigned int			count;
	bool					allMode, indivMode;
	unsigned int			maxNum;

 protected:
	// 新パーティクル生成 
	fk_Particle *	newParticle(void);
	fk_Particle *	newParticle(const fk_Vector &);
	fk_Particle *	newParticle(double, double, double);

	// パーティクル削除 
	bool			removeParticle(fk_Particle *);
	bool			removeParticle(int);

	// 年齢取得 
	unsigned int	getCount(void) const;

	// ID からパーティクルを取得 
	fk_Particle *	getParticle(int) const;

	// 全体パーティクル取得 
	fk_Particle *	getNextParticle(fk_Particle *) const;

	// パーティクル個数取得 
	unsigned int	getParticleNum(void) const;

	// パーティクル最大個数設定取得 
	void			setMaxSize(unsigned int);
	unsigned int	getMaxSize(void) const;

	// カラーパレット設定 
	void			setColorPalette(int, const fk_Color &);
	void			setColorPalette(int, float, float, float);
	void			setColorPalette(int, double, double, double);

	// 個別初期化用仮想関数 
	virtual void	genMethod(fk_Particle *);

	// 全体動作用仮想関数 
	virtual void	allMethod(void);

	// 個別動作用仮想関数 
	virtual void	indivMethod(fk_Particle *);

	// 全体動作モード設定取得 
	void			setAllMode(bool);
	bool			getAllMode(void) const;

	// 個別動作モード設定取得 
	void			setIndivMode(bool);
	bool			getIndivMode(void) const;

 public:
	fk_ParticleSet(unsigned int = 0);
	virtual ~fk_ParticleSet();

	// 動作 
	void			handle(void);

	// シーン設定用 
	fk_Shape *		getShape(void) const;
};

#endif // !__FK_PARTICLE_HEADER__
